// PROGRAM: backdsat.cc
// AUTHOR: Koen van der Hauw
// Implementation of DSatur together with backtracking 
// Implmented to be used for Master's Thesis for comparison
// with the EA genea.

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <MLCG.h>
#include <sys/time.h>
#include <math.h>
#include <string.h>

// Put values for n,p and s in function main() and function MakeReport()

#define NR_OF_COLORS 3
#define MAXNODES 1300   // Above 1500 nodes problems start to occur...

#define MAXRUNS 100     // MAXRUNS just should be bigger than NROFRUNS
#define NROFRUNS 25

#define MAXSTEPS 1000000
#define GRAPHNAME "eq"
#define GRAPHDIR "../problems/"
#define RESULTDIR "../results/"
#define FILELABEL "timecomplpn1250"
#define MESCERTAINTY 0.9       // Certainty of Minimal Expected nr of Evals 
#define MAXPATHLENGTH 250

class Link {
public:
  int nodenr;
  Link *nextptr;
};

class GraphClass {
public:
  int nrofnodes;

  char GraphFilename[MAXPATHLENGTH];
  Link* graph[MAXNODES];            /* edges of graph in adjacency list */
  int nrofneighbors[MAXNODES];
  void CreateGraph(int, double, int);
  void RemoveGraph();
  void DeleteAdjlist();
  void PrintAdjList();
  void ReadGraph(int&);  
};


void GraphClass::DeleteAdjlist() {

  int i;
  Link *temp;

  for (i = 0; i < nrofnodes; i++)
    while (graph[i]) {
      temp = graph[i]->nextptr;
      delete graph[i];
      graph[i] = temp;
    }

} /* GraphClass::DeleteAdjlist() */


void GraphClass::PrintAdjList() {
  
  int i;
  Link *temp;
 
  for (i = 0; i < nrofnodes; i++) {
    cout << "buren van node " << i << ": ";
    temp = graph[i];
    while (temp) {
      cout << temp->nodenr << " ";
      temp = temp->nextptr;
    }
    cout << endl;
  }

} /* GraphClass::DeleteAdjList() */

void GraphClass::CreateGraph(int n, double p, int s) {
  
  int i, l;
  char string[MAXPATHLENGTH];
  ofstream output;
  const int DECIMALS = 3;
    
  strcpy(GraphFilename, GRAPHNAME);
  l = strlen(GraphFilename);
  GraphFilename[l++] = 'n';
  for (i = (int) log10((double) n); i >= 0; i--) 
    GraphFilename[l++] = '0' + (n % ((int) pow(10,i+1))) / ((int) pow(10, i));
  GraphFilename[l++] = 'p';
  if (p < 1.0) 
    GraphFilename[l++] = '0';
  else
    GraphFilename[l++] = '1';
  GraphFilename[l++] = '.';

  for (i = 1; i <= DECIMALS; i++) 
    GraphFilename[l++] = '0' + (((int)(pow(10, i) * p + pow(10,-DECIMALS-1))) % 10);
  GraphFilename[l++] = 's';
  GraphFilename[l++] = '0' + s / 10;
  GraphFilename[l++] = '0' + s % 10;
  GraphFilename[l++] = (char) NULL;

  strcpy(string, GRAPHDIR);
  strcat(string, GraphFilename);
  strcat(string,"ingen");

  output.open(string);
  output << 0 << endl;      // ascii code
  output << s << endl;      // seednr s
  output << 2 << endl;      // equipartite
  output << n << endl;      // nr of nodes
  output << 3 << endl;      // 3-colorings
  output << 1 << endl;      // IID-edge probability
  output << p << endl;      // edge probabilty p
  output << 0 << endl;      // no cheats
  output.close();

  strcpy(string, "generator ");
  strcat(string, GRAPHDIR);
  strcat (string, GraphFilename);
  strcat(string, " < ");
  strcat(string, GRAPHDIR);
  strcat(string, GraphFilename);
  strcat(string, "ingen");
  strcat(string, " > ");
  strcat(string, GRAPHDIR);
  strcat(string, GraphFilename);
  strcat(string, "NULL");
  system(string);
 
} /* GraphClass::CreateGraph() */

void GraphClass::RemoveGraph() {

  char string[MAXPATHLENGTH];

  strcpy(string, "rm ");
  strcat(string, GRAPHDIR);
  strcat(string, GraphFilename);
  system(string);
  strcpy(string, "rm ");
  strcat(string, GRAPHDIR);
  strcat(string, GraphFilename);
  strcat(string, "ingen");
  system(string);

  strcpy(string, "rm ");
  strcat(string, GRAPHDIR);
  strcat(string, GraphFilename);
  strcat(string, "NULL");
  system(string);

  DeleteAdjlist();

} /* GraphClass::RemoveGraph() */


void GraphClass::ReadGraph(int &nrofgenes) {

  const int MAXLEN = 250;
  ifstream input;
  int n1, n2, i;
  char kar, string[MAXLEN];
  Link  *temp;

  strcpy (string, GRAPHDIR);
  strcat (string, GraphFilename);
  input.open(string);

  cout << "Graph \""<< string << "\" ";
  /* Find line starting with p, which includes the nr of nodes */
  input >> kar;
  while (kar != 'p') {
    input.getline(string, MAXLEN);
    input >> kar;
  }
  
  input >> string;
  input >> nrofnodes;
  nrofgenes = nrofnodes;
  input.getline(string, MAXLEN);

  /* Initialize edges of graph at non-existent */
  for (i = 0; i < nrofgenes; i++) {
    graph[i] = NULL;
    nrofneighbors[i] = 0;
  }
  
  /* Find first line starting with an 'e' to get the edges */ 
  input >> kar;
  while (kar != 'e') {
    input.getline(string, MAXLEN);
    input >> kar;
  }
  while (!input.eof()) {
    input >> n1 >> n2;
    /* cout << "Edge (" << n1 << "," << n2 << ")" << endl; */
    temp = graph[n1-1];
    graph[n1-1] = new Link;
    graph[n1-1]->nodenr = n2-1;
    graph[n1-1]->nextptr = temp;
    temp = graph[n2-1];
    graph[n2-1] = new Link;
    graph[n2-1]->nodenr = n1-1;
    graph[n2-1]->nextptr = temp;
    input >> kar;
  }

  input.close();

  cout << "read." << endl;
  cout << nrofgenes << " nodes" << endl;
} /* GraphClass::ReadGraph() */


// Application Specific Functions and Definitions
/* Global variables */
int nrofnodes;         /* Number of nodes for current graph problem */
GraphClass graph;
MLCG rnd;   /* class to generate random numbers */


/* Gives time in seconds to use as a seed number */
int getsec() {
  
  struct timeval tp;

  gettimeofday(&tp);
  return tp.tv_usec;
  
} /* getsec() */


clock_t starttime,
     stoptime;

void StartTimer() {
 
  starttime = clock();

} /* StartTimer() */

void StopTimer() {

  stoptime = clock();

} /* StopTimer() */

double GetRunTime() {

  return ((double) (stoptime - starttime) / CLOCKS_PER_SEC);

} /* RetRunTime() */

class HeapInfo {
public:
  int degree;
  int satdegree;
  bool avail[NR_OF_COLORS];
  int color;
  int heapindex;
};

class DSatHeap {
public:
  int heapsize;
  int nodenr[MAXNODES+1];          /* number of node in graph */
  HeapInfo info[MAXNODES];         /* info with index nr of node in graph */
  void Initialize();
  void MakeHeap();
  void Heapify(int);
  int DeleteMax();
  void Sift(int);
  bool Bigger(int, int);
  void PrintHeap();
  void Copy(DSatHeap&);
private:
  int GetNrOfMax(int);
  int GetRndMax(int, int&, int);
};

void DSatHeap::Initialize() {

  int i, k;
  Link *ptr;

  /* Initialize Heap */
  for (i = 0; i < nrofnodes; i++) {
    nodenr[i+1] = i;
    info[i].satdegree = 0;
    info[i].degree = 0;
    for (ptr = graph.graph[i]; ptr; ptr = ptr->nextptr)
      info[i].degree++;
    for (k = 0; k < NR_OF_COLORS; k++)
      info[i].avail[k] = true;
    info[i].heapindex = i+1;
    info[i].color = -1;
  }

} /* DSatHeap::Initialize() */

/* returns true if (heap)node i is bigger than (heap)node j   */
bool DSatHeap::Bigger(int i, int j) {

  if (info[nodenr[i]].satdegree > info[nodenr[j]].satdegree)
    return true;
  else
    if (info[nodenr[i]].satdegree == info[nodenr[j]].satdegree)
      if (info[nodenr[i]].degree > info[nodenr[j]].degree)
        return true;
      else
        return false;
    else
      return false;

} /* DSatHeap::Bigger() */

void DSatHeap::MakeHeap() {

  int i;

  heapsize = nrofnodes;
  for (i = heapsize/2; i >= 1; i--)
    Heapify(i);

} /* DSatHeap::MakeHeap() */

void DSatHeap::Heapify(int i) {
  
  int l;
  int r;
  int largest;
  int temp;
  bool finished;

  finished = false;
  do {
    l = 2*i;
    r = 2*i+1;
    if (l <= heapsize && Bigger(l,i))
      largest = l;
    else
      largest = i;
    if (r <= heapsize && Bigger(r,largest))
      largest = r;
    if (largest != i) {
      temp = nodenr[i]; 
      nodenr[i] = nodenr[largest];
      nodenr[largest] = temp;
      info[nodenr[i]].heapindex = i;
      info[nodenr[largest]].heapindex = largest;
      i = largest;
    }
    else
      finished = true;
  } while (!finished);
   
} /* DSatHeap::Heapify() */

void DSatHeap::PrintHeap() {

  int i;

  for (i = 1; i <= heapsize; i++) 
    cout << nodenr[i] << "(" << info[nodenr[i]].satdegree << "," 
         << info[nodenr[i]].degree << ")" << " ";
  cout << endl;

} /* DSatHeap::PrintHeap() */

/* return number of maxima in subheap starting at node i */
int DSatHeap::GetNrOfMax(int i) {

  int nrofmax = 0;

  if (2*i <= heapsize)
    if (!Bigger(i, 2*i)) 
      nrofmax += GetNrOfMax(2*i);
  if (2*i+1 <= heapsize)
    if (!Bigger(i, 2*i+1))
      nrofmax += GetNrOfMax(2*i+1);
  return (1 + nrofmax);

} /* DSatHeap::GetNrOfMax() */

/* Return index of 'rndmax'-nd maximum in subheap starting at node i, */
/* assuming 'maxnr' maxima have been found already.                   */
/* after returning 'maxnr' will be the nr. of maxima in the subheap   */
/* returning of 0 means the searched maximum is not is this subheap   */
int DSatHeap::GetRndMax(int i, int &maxnr, int rndmax) {
  
  int max;

  maxnr++;

  if (maxnr == rndmax)
    return i;
 
  if (2*i <= heapsize)
    if (!Bigger(i, 2*i))
      if (max = GetRndMax(2*i, maxnr, rndmax))
        return max;

  if (2*i+1 <= heapsize)
    if (!Bigger(i, 2*i+1))
      if (max = GetRndMax(2*i+1, maxnr, rndmax))
        return max;
  
  return 0;

} /* DSatHeap::GetRndMax() */

/* Return maximum element of heap. If there are several maximal */
/* elements, randomly select one.                               */
int DSatHeap::DeleteMax() {
  
  int i = 0,  /* necessary for recursive function GetRndMax */
      maxindex,   /* index of chosen maximum node */
      max,        /* nodenr in graph of chosen maximum element */
      maxnr;
  
  /* Get number of possible candidates */
  maxnr = GetNrOfMax(1);
  
  if (maxnr > 1)
    maxnr = (rnd.asLong() % maxnr) + 1;
  maxindex = GetRndMax(1, i, maxnr);
  
  max = nodenr[maxindex];
  nodenr[maxindex] = nodenr[heapsize];
  heapsize--;
  info[nodenr[maxindex]].heapindex = maxindex;
  
  /* preserve heap-characteristic */
  Sift(maxindex);

  return max; 

} /* DSatHeap::DeleteMax() */

/* move node with index i up in the heap to right position */
void DSatHeap::Sift(int i) {

  int p;
  int temp;
  int startindex = i;

  p = i/2;
  temp = nodenr[i];
  while (p >= 1 && Bigger(i,p)) {
    nodenr[i] = nodenr[p];
    nodenr[p] = temp;
    info[temp].heapindex = p;
    info[nodenr[i]].heapindex = i;
    i = p; 
    p = i/2;
  }

  if (i == startindex)
    Heapify(i);
    
} /* DSatHeap::Sift() */


void DSatHeap::Copy(DSatHeap &h) {

  int i, k;

  heapsize = h.heapsize;
  for (i = 1; i <= heapsize; i++) 
    nodenr[i] = h.nodenr[i];
  for (i = 0; i < nrofnodes; i++) {
    info[i].satdegree = h.info[i].satdegree;
    info[i].degree = h.info[i].degree;
    info[i].color = h.info[i].color;
    for (k = 0; k < NR_OF_COLORS; k++)
      info[i].avail[k] = h.info[i].avail[k];
    info[i].heapindex = h.info[i].heapindex;
  }

} /* DSatHeap::Copy() */

class DSat {
public:
  long steps;
  int minheapsize;
  void Initialize();
  bool BackTrack(DSatHeap);
  void FindSolution();
};

void DSat::Initialize() {

  steps = 0;
  minheapsize = nrofnodes;

} /* DSat::Initialize() */

bool DSat::BackTrack(DSatHeap h) {

  int k;
  Link *ptr;
  int x;
  DSatHeap oldheap;

  // if (steps % 10000 == 0) 
  //  cout << "dsatstep: " << steps << ", heapsize: " << h.heapsize << endl;
    
   if (steps > MAXSTEPS) {
     return false;
    // cout << steps << ", ";
   }

  steps++;
//  if (steps % 10000 == 0)
  //  cout << steps << "steps done..." << endl;

  if (h.heapsize == 0) {
      /*  cout << "Solution Found: ";
    for (k = 0; k < nrofnodes; k++)
      cout << h.info[k].color;
      cout << endl; */
    return true;
  }
  else {
    x = h.DeleteMax();
    if (h.heapsize < minheapsize)
      minheapsize = h.heapsize;
    oldheap.Copy(h);
    k = 0;
    while (k < NR_OF_COLORS) {
      /* Select next possible color for node x */
      while (k < NR_OF_COLORS && !(h.info[x].avail[k])) 
        k++;
      if (k < NR_OF_COLORS) {      
        h.info[x].color = k;
        ptr = graph.graph[x];
        while (ptr) {
          if (h.info[ptr->nodenr].color == -1) {
            if (h.info[ptr->nodenr].avail[h.info[x].color]) {
              h.info[ptr->nodenr].avail[h.info[x].color] = false;
              h.info[ptr->nodenr].satdegree++;
            }
            h.info[ptr->nodenr].degree--;
            h.Sift(h.info[ptr->nodenr].heapindex);
          }  
          ptr = ptr->nextptr;
        }
        if (BackTrack(h)) 
          return true;
        else
          if (steps > MAXSTEPS)
            return false;
      } 
      h.Copy(oldheap);
      k++;
    }
    return false;
  }       


} /* DSat::BackTrack() */

void DSat::FindSolution() {

  DSatHeap h;

  h.Initialize();
  h.MakeHeap();  

  BackTrack(h);

} /* DSat::FindSolution() */




void MakeReport(int n, int s) {

  int i, r;
  double p;
  DSat dsat;
  int nrofsuccesses;
  long mcnsum;
  long mcnsqrsum;
  double sr;
  double sssum;
  double sssqrsum;
  long stepsneed[MAXRUNS];
  double steptimesum;
  long bestexpruns;
  long bestexpsteps;
  long expruns;
  long expsteps;
  ofstream reportoutput;
  ofstream gnuplotoutput;

  char reportname[MAXPATHLENGTH];

  strcpy(reportname, RESULTDIR);
  strcat(reportname, "dsat");
  strcat(reportname, FILELABEL);
  strcat(reportname, ".report");
  reportoutput.open(reportname); 

  strcpy(reportname, RESULTDIR);
  strcat(reportname, "dsat");
  strcat(reportname, FILELABEL);
  strcat(reportname, ".gnu");
  gnuplotoutput.open(reportname); 
 
  gnuplotoutput << "#Number of nodes: " << n << endl
                << "#equipartite graphs with seed " << s << endl
                << "#Number of runs: " << NROFRUNS << endl
                << "#Maximum number of Steps: " << MAXSTEPS << endl;
  reportoutput << "Number of runs: " << NROFRUNS << endl
               << "Maximum number of Steps: " << MAXSTEPS << endl;
  reportoutput << endl << "-------------------------------RESULTS PER GRAPH---------------------------" << endl;

  reportoutput << "SR = Succes Rate" << endl
     << "AMCN = Avg. Maximum of colored nodes (Avg. over all runs with NO solutions)" << endl
     << "SMCN = Standard Deviation of SMCN" << endl
     << "ASS = Avg. nr. of steps done when GA is stopped (Avg. over all runs)" << endl
     << "SSS = Standard Deviation of ASS" << endl 
     << "MES = Minimal Expected nr of STEPS to find a solution with " 
     << MESCERTAINTY * 100 << "\% certainty" << endl 
     << "ATS = Average Time per Step (in seconds/1000 steps)" << endl
     << "MET = Minimal Expected Time to find a solution with "
     << MESCERTAINTY * 100 << "\% certainty" << endl
     << "-------------------------------------------------------------------------------"<< endl;
  reportoutput << "Graph:          SR, AMCN,  SMCN,    ASS,     SSS,     MES,    ATE,     MET" << endl
     << "-------------------------------------------------------------------------------"<< endl;


  for (p = 8.0/n; p < 1.0; p += 1) {
    cout << "p = " << p << endl;
    graph.CreateGraph(n, p, s);
    graph.ReadGraph(nrofnodes);

    reportoutput << graph.GraphFilename << ": "; 
  
    nrofsuccesses = 0;
    mcnsum = 0;
    mcnsqrsum = 0;
    sssum = 0;
    sssqrsum = 0;
    steptimesum = 0;

    for (r = 0; r < NROFRUNS; r++) {
      rnd.seed1(0);
      rnd.seed2(r);

      StartTimer(); 

      /* Run Brelaz' DSatur */
      dsat.Initialize();
      dsat.FindSolution();

      StopTimer();
      steptimesum += ((double) GetRunTime() / dsat.steps);
      if (dsat.minheapsize == 0) {
        /* store in sorted list when solution was found */
	cout << "solution found for run " << r << endl;
        i = nrofsuccesses;
        while ((i > 0) && (dsat.steps < stepsneed[i-1])) {
          stepsneed[i] = stepsneed[i-1];
          i--;
        }
        stepsneed[i] = dsat.steps;
        nrofsuccesses++;
        sssum += dsat.steps;
        sssqrsum += ((double) dsat.steps * dsat.steps);
        
      }
      else {
	cout << "no solution found for run " << r  << endl;
        mcnsum += dsat.minheapsize;
        mcnsqrsum += (dsat.minheapsize * dsat.minheapsize);
      }
    }

    bestexpsteps = -1;
    for (i = 0; i < nrofsuccesses; i++) {
      sr = (double)(i+1)/(double) NROFRUNS;
      if (sr < 1.0)
        expruns = (int) ceil (log(1.0 - MESCERTAINTY) / 
                              log(1.0 - sr));
      else
        expruns = 1;
      expsteps = expruns * stepsneed[i];
      if ((expsteps < bestexpsteps) || (bestexpsteps < 0)) {
        bestexpsteps = expsteps;
        bestexpruns = expruns;
      }
    }  

    gnuplotoutput << p << " " << (double) nrofsuccesses / NROFRUNS << endl;
    reportoutput << (double) nrofsuccesses / NROFRUNS << ", " 
                 << (double) mcnsum / (NROFRUNS - nrofsuccesses) << ", "
                 << sqrt(mcnsqrsum-mcnsum*mcnsum/(double)(NROFRUNS-nrofsuccesses))/(NROFRUNS - nrofsuccesses - 1) << ", " 
                 << (double) sssum / nrofsuccesses << ", " 
                 << sqrt(sssqrsum-sssum*sssum/(double)nrofsuccesses)/(nrofsuccesses-1)
                 << ", " << bestexpsteps << ", "
                 << (steptimesum / NROFRUNS) * 1000 << ", "
                 << (bestexpsteps * (steptimesum / NROFRUNS)) << endl;

    graph.RemoveGraph();
  }
    
  reportoutput.close();   
  gnuplotoutput.close();
   
} /* RunDsatur() */



main() {

  MakeReport(1250, 5);

}


